import os
from typing import Dict, Optional
import copy
from kubernetes.client import models as k8s

class AstrafyTrainingEnvironment:
    """
    A class to manage environment-specific configurations for project, dataset, and service account.
    Allows overriding default values via constructor parameters.
    """

    def __init__(self, environment: str = None,package: str = None, configs: Optional[Dict[str, Dict[str, str]]] = None):
        """
        Initializes the AstrafyTrainingEnvironment with the given environment and optional overrides for project, dataset, and service account.
        If no environment is provided, it attempts to determine it from the OS environment variable 'ENVIRONMENT'.

        Args:
            environment (str, optional): The environment (e.g., 'dev', 'prd'). Defaults to None.
            configs (Dict[str, Dict[str, str]], optional): A dictionary containing environment-specific configurations. Defaults to None.
        """
        self.environment = environment or os.getenv('ENVIRONMENT')
        self.package = package
        env_vars = [
            self.set_env_vars(name="STG_PROJECT_ID", key="STG_PROJECT_ID"),
            self.set_env_vars(name="LZ_PROJECT_ID", key="LZ_PROJECT_ID"),
            self.set_env_vars(name="DM_PROJECT_ID", key="DM_PROJECT_ID"),
            self.set_env_vars(name="DW_PROJECT_ID", key="DW_PROJECT_ID"),
            self.set_env_vars(name="PACKAGE", value=self.package),
            self.set_env_vars(name="ENV", value=self.environment),
        ]
        
        self.env_vars = env_vars

        if not self.environment:
            raise ValueError("Environment not specified. Please provide an environment or set the 'ENVIRONMENT' environment variable.")

        self.config = self._load_config(configs)


    def _load_config(self, configs: Optional[Dict[str, Dict[str, str]]] = None) -> Dict:
        """
        Loads the configuration based on the environment, using provided configs if available.

        Returns:
            Dict: A dictionary containing the configuration for the specified environment.

        Raises:
            ValueError: If the environment is not supported.
        """
        # Define default configurations for different environments
        default_config = {
            "dev": {
                "project": "dev-project-id",
                "dataset": "dev_dataset",
                "service_account": "dev-service-account@dev-project-id.iam.gserviceaccount.com"
            },
            "prd": {
                "project": "prd-project-id",
                "dataset": "prd_dataset",
                "service_account": "prd-service-account@prd-project-id.iam.gserviceaccount.com"
            }
        }

        if configs and self.environment in configs:
            # Merge the provided configs with the default configs
            config = default_config.get(self.environment, {}).copy()  # Start with a copy of the default config
            config.update(configs[self.environment])  # Update with values from configs
        else:
            config = default_config.get(self.environment, {})

        if not config:
            raise ValueError(f"Unsupported environment: {self.environment}. Supported environments are: {', '.join(default_config.keys())}")

        return config
    def set_env_vars(self,name: str, value=None, key=None):
        if value:
            env_var = k8s.V1EnvVar(name=name, value=value)
            return env_var
        if key:
            env_var = k8s.V1EnvVar(name=name, value_from=k8s.V1EnvVarSource(
                config_map_key_ref=k8s.V1ConfigMapKeySelector(name="internal-data-dbt-config-map", key=key)))
            return env_var 
        
    def add_env_vars(self,additional_vars=None):
        """
        Set up environment variables for a given data product and environment,
        with the option to add additional variables.
        """

        # Extend with additional environment variables specific to the DAG
        if additional_vars:
            config_vars = copy.deepcopy(self.env_vars)
            print(additional_vars)
            config_vars.extend([
                self.set_env_vars(name=var_name, value=var_value) for var_name, var_value in additional_vars.items()
            ])

            self.env_vars  = copy.deepcopy(config_vars)
        
        return self.env_vars   
    def get_project(self) -> str:
        """
        Returns the project ID for the current environment.

        Returns:
            str: The project ID.
        """
        return self.config["project"]

    def get_dataset(self) -> str:
        """
        Returns the dataset ID for the current environment.

        Returns:
            str: The dataset ID.
        """
        return self.config["dataset"]

    def get_service_account(self) -> str:
        """
        Returns the service account for the current environment.

        Returns:
            str: The service account for the current environment.
        """
        return self.config["service_account"]




from airflow.contrib.operators.slack_webhook_operator import SlackWebhookOperator
from airflow.hooks.base_hook import BaseHook
from airflow.configuration import conf


SLACK_CONN_ID = "slack_airflow"


def task_fail_slack_alert(context):
    webserver_url = conf.get('webserver', 'base_url')
    dag_id = context.get("task_instance").dag_id
    task_id = context.get("task_instance").task_id
    run_id = context.get("task_instance").run_id
    execution_date = context.get("execution_date")
    dag_url = f"{webserver_url}/dags/{dag_id}/grid?execution_date={execution_date.isoformat()}&tab=graph&dag_run_id={run_id}"

    slack_msg = """
            :red_circle: Task Failed.
            *Task*: {task}  
            *Dag*: {dag} 
            *Execution Time*: {exec_date}
            *Dag URL*: <{url}|View in Airflow>
            """.format(
        task=task_id,
        dag=dag_id,
        exec_date=execution_date,
        url=dag_url
    )
    failed_alert = SlackWebhookOperator(
        task_id="slack_fail_alert", 
        slack_webhook_conn_id=SLACK_CONN_ID,
        message=slack_msg,
    )
    return failed_alert.execute(context=context)



from datetime import datetime, timedelta
from airflow.models.param import Param

def default_dag(data_product, ENV, schedule= '0 10 * * *',start_date=datetime(2022, 10, 10, 1, 0),**kwargs):
    dag = {
        "dag_id": f'{data_product}-k8s-dag-{ENV}',
        "schedule_interval": schedule,
        "description": f'Dag for the {data_product} data product - {ENV}',
        "max_active_tasks": 10,
        "catchup": False,
        "is_paused_upon_creation": True,
        "tags": ['k8s', data_product],
        "max_active_runs": 1,
        "dagrun_timeout": timedelta(seconds=36000),
        "default_view": 'grid',
        "orientation": 'LR',
        "sla_miss_callback": None,
        "params": {"is_full_refresh": Param(False, type="boolean")},
        "start_date": start_date,
        "doc_md": __doc__,
    }
    
    dag.update(kwargs)
    
    return dag


from airflow import DAG
from airflow.models import BaseOperator
from airflow.models.dag import DagContext
from airflow.providers.google.cloud.operators.pubsub import PubSubPublishMessageOperator
from airflow.utils.trigger_rule import TriggerRule


def build_pubsub_message(message_dict) -> dict:

    return {
        'data': bytes(str(message_dict), encoding='utf-8')
    }


def notification_job_for_table(
        task_id: str,
        message_dict,
        downstream_project,
        topic,
        impersonation_chain="",
        trigger_rule: TriggerRule = TriggerRule.ALL_SUCCESS,
        dag: DAG = None,) -> BaseOperator:
    """
        Creates an Airflow job to send a Pub/Sub notification message for the list of given tables
        or log a message based on the environment.

        Parameters
        ----------
        task_id
            The ID of the Airflow task
        task_group
            The group Task of Airflow to group the tasks
        table
            The modified table
        flow
            The flow concerned by the notification
        dataset
            The dataset concerned by the notification
        mode
            The mode for the job Can be 'incremental' or 'full'
        downstream_project
            The downstream project where topic is
        topic
            The topic name
        env
            The environment identifier 
        trigger_rule
            Trigger rule of the messaging system based on your logic
        dag
            The dag in which the operator must be instantiated
        
        
        

        Returns
        -------
        operator
            Returns an PubSubOperator that will publish the normalized message

        Raises
        ------
        ValueError
            If an unsupported mode is provided.

        """
    dag = dag or DagContext.get_current_dag()

    job= PubSubPublishMessageOperator(
        dag=dag,
        task_id=task_id,
        trigger_rule=trigger_rule,
        project_id=downstream_project,
        topic=topic,
        impersonation_chain=impersonation_chain,
        messages=[build_pubsub_message(message_dict)])
    
    return job



import random
import datetime
import string
from google.cloud import bigquery

class BigQueryClient:
    """
    BigQuery Client 
    """

    
    def __init__(self, bq_project,bq_dataset):

        self.BQ_PROJECT = bq_project
        self.BQ_DATASET = bq_dataset
        self.client = self.create_client()


    # Create a client
    def create_client(self):
        try:
            return bigquery.Client(
                project=self.BQ_PROJECT
            )
        except Exception as e:
            print(f'Error creating client: {e}')
            return None
    
    def insert_data_to_bq(self, data, table_name):
        """
        Insert data into BigQuery table.
        """
        if not self.client:
            print("BigQuery client is not initialized.")
            return

        table_id = f"{self.BQ_PROJECT}.{self.BQ_DATASET}.{table_name}"
        errors = self.client.insert_rows_json(table_id, data)

        if errors:
            print("Error inserting into {table_name}: {errors}")
        else:
            print(f"Successfully loaded {len(data)} rows into {table_name}")

# Helper Functions
def random_string(length):
    return "".join(random.choices(string.ascii_uppercase, k=length))

def random_date(start_date, end_date):
    delta = end_date - start_date
    random_days = random.randint(0, delta.days)
    return start_date + datetime.timedelta(days=random_days)

def generate_customer(num_rows):
    return [
        {
            "id": f"CUST{str(i).zfill(4)}",
            "name": f"{random_string(5)} {random_string(7)}",
            "email": f"{random_string(5).lower()}@example.com",
            "country": random.choice(["US", "EU", "UK"]),
            "signup_date": random_date(datetime.date(2020, 1, 1), datetime.date(2023, 10, 1)).isoformat(),
            "ingested_date": datetime.date.today().isoformat(),
        }
        for i in range(1, num_rows + 1)
    ]

def generate_products(num_rows):
    categories = ["Electronics", "Clothing", "Home", "Sports", "Books"]
    return [
        {
            "id": f"PROD{str(i).zfill(4)}",
            "name": f"Product {random_string(5)}",
            "category": random.choice(categories),
            "price": round(random.uniform(10, 500), 2),
            "in_stock": random.choice([True, False]),
            "ingested_date": datetime.date.today().isoformat(),
        }
        for i in range(1, num_rows + 1)
    ]

def generate_transactions(num_rows, customer_ids, product_ids):
    return [
        {
            "id": f"TXN{str(i).zfill(4)}",
            "customer_id": random.choice(customer_ids),
            "product_id": random.choice(product_ids),
            "amount": round(random.uniform(20, 500), 2),
            "currency": random.choice(["USD", "EUR", "GBP"]),
            "transaction_date": random_date(datetime.date(2023, 1, 1), datetime.date(2023, 10, 31)).isoformat(),
            "ingested_date": datetime.date.today().isoformat(),
        }
        for i in range(1, num_rows + 1)
    ]



def generate_and_load(rows_per_table, bq_project,bq_dataset,data_product):
  
    bq_client = BigQueryClient(bq_project,bq_dataset)
    # Generate customer
    if data_product.lower() == "customer":
        customer = generate_customer(rows_per_table)
        bq_client.insert_data_to_bq(customer, "customers")
        return "customers_load"
    elif data_product.lower() == "product":
    # Products
        products = generate_products(rows_per_table)
        bq_client.insert_data_to_bq(products, "products")
        return "products_load"
    elif data_product.lower() == "transaction":
    # Transactions
        transactions = generate_transactions(
            rows_per_table,
            [c["id"] for c in customer],
            [p["id"] for p in products],
        )
        # Split transactions based on currency
        uk_transactions = [t for t in transactions if t["currency"] == "GBP"]
        us_transactions = [t for t in transactions if t["currency"] == "USD"]
        eu_transactions = [t for t in transactions if t["currency"] == "EUR"]
        
        # Insert transactions into relevant tables
        bq_client.insert_data_to_bq(uk_transactions, "transactions_uk")
        bq_client.insert_data_to_bq(us_transactions, "transactions_us")
        bq_client.insert_data_to_bq(eu_transactions, "transactions_eu")
        return "transactions_load"
    else:
        raise ValueError(f"Invalid Data product: {data_product}. Use 'customers' or 'products' or 'transactions'.")



from os import environ
from kubernetes.client import models as k8s
from airflow.providers.cncf.kubernetes.operators.pod import KubernetesPodOperator
from airflow.models import Variable
from airflow import DAG
from airflow.models import Variable
from airflow.models.param import Param
from datetime import datetime, timedelta

class ConfigVars:

    def __init__(self):
        COMMON_GITLAB_PAT = Variable.get("COMMON_GITLAB_PAT")
        env_vars = [
            set_env_vars(name="STG_PROJECT_ID", key="STG_PROJECT_ID"),
            set_env_vars(name="LZ_PROJECT_ID", key="LZ_PROJECT_ID"),
            set_env_vars(name="DM_PROJECT_ID", key="DM_PROJECT_ID"),
            set_env_vars(name="DW_PROJECT_ID", key="DW_PROJECT_ID"),
            set_env_vars(name="COMMON_GITLAB_PAT", value=COMMON_GITLAB_PAT)
        ]
        self.env_vars = env_vars


class DbtRunDefaults:
    image_pull_policy = "Always"
    cmds = ["/bin/bash", "-c"]

    get_logs = True
    service_account_name = 'dbt-ksa'
    namespace = "dbt"
    affinity = {'nodeAffinity': {'requiredDuringSchedulingIgnoredDuringExecution':
        {'nodeSelectorTerms': [
            {'matchExpressions': [
                {'key': 'node_pool',
                 'operator': 'In',
                 'values': ['dbt']
                 }
            ]
            }
        ]
        }
    }
    }

    def __init__(self, data_product, environment):
        self.dbt_target_gke = f'{data_product}-{environment}-ci-airflow'
        self.dbt_target = f""" --target="{self.dbt_target_gke}" """
        self.run_id = '{{ run_id }}'
        self.dag_ts = '{{ ts }}'
        self.dbt_profile = f""" --profiles-dir=/app/artifacts"""
        self.dbt_vars = f""" --vars '{{"dag_id": "{self.run_id}" , "dag_ts": "{self.dag_ts}", "orchestrator": "Airflow", 
        "job_run_id": "{self.run_id}"}}'"""
        self.dbt_args = f"""{self.dbt_vars}{self.dbt_target}{self.dbt_profile}"""
        self.docker_image = "europe-west1-docker.pkg.dev/prj-astrafy-artifacts/mds"

def default_dag(data_product, ENV, **kwargs):
    dag = {
        "dag_id": f'{data_product}-k8s-dag-{ENV}',
        "schedule_interval": '0 10 * * *',
        "description": f'Dag for the {data_product} data product - {ENV}',
        "max_active_tasks": 10,
        "catchup": False,
        "is_paused_upon_creation": True,
        "tags": ['k8s', data_product],
        "max_active_runs": 1,
        "dagrun_timeout": timedelta(seconds=36000),
        "default_view": 'grid',
        "orientation": 'LR',
        "sla_miss_callback": None,
        "params": {"is_full_refresh": Param(False, type="boolean")},
        "start_date": datetime(2022, 10, 10, 1, 0),
        "doc_md": __doc__,
    }
    
    dag.update(kwargs)
    
    return dag


def cd_to_local_package(package: str) -> str:
    return f"cd /app/data_products/{package}/"


def set_env_vars(name: str, value=None, key=None):
    if value:
        env_var = k8s.V1EnvVar(name=name, value=value)
        return env_var
    if key:
        env_var = k8s.V1EnvVar(name=name, value_from=k8s.V1EnvVarSource(
            config_map_key_ref=k8s.V1ConfigMapKeySelector(name="internal-data-dbt-config-map", key=key)))
        return env_var

def setup_env_vars(data_product, environment, additional_vars=None):
    """
    Set up environment variables for a given data product and environment,
    with the option to add additional variables.
    """
    config_vars = ConfigVars()

    # Extend with additional environment variables specific to the DAG
    additional_env_vars = [
        set_env_vars(name="DATA_PRODUCT", value=data_product),
        set_env_vars(name="ENV", value=environment)
    ]

    if additional_vars:
        additional_env_vars.extend([
            set_env_vars(name=var_name, value=var_value) for var_name, var_value in additional_vars.items()
        ])

    config_vars.env_vars.extend(additional_env_vars)
    return config_vars.env_vars

def construct_dbt_run_command(dbt_defaults, selector, mode):
    """
    Construct a DBT command using the DbtRunDefaults settings, a selector, and a specified mode.
    """
    dbt_selector = f"--selector {selector}"
    return f"dbt run {dbt_defaults.dbt_args} {dbt_selector} {mode};exit $?"

def create_dbt_task(task_id, name, command, dbt_defaults, env_vars, data_product):
    """
    Create a KubernetesPodOperator task for running DBT commands.
    """
    DP_VERSION = Variable.get(f"DP_{data_product.upper()}_VERSION")
    return KubernetesPodOperator(
        task_id=task_id,
        name=name,
        image=f"{dbt_defaults.docker_image}/dbt/{data_product}:{DP_VERSION}",
        cmds=dbt_defaults.cmds,
        arguments=[command],
        image_pull_policy=dbt_defaults.image_pull_policy,
        namespace=dbt_defaults.namespace,
        get_logs=dbt_defaults.get_logs,
        service_account_name=dbt_defaults.service_account_name,
        affinity=dbt_defaults.affinity,
        env_vars=env_vars,
    )


import copy
from airflow.models import Variable
from dataclasses import dataclass
from kubernetes.client import models as k8s

@dataclass
class AstrafyEnvironment:
    """
    This class contains the default environment variables available during the DAG execution.
    """
   
    def __init__(self, data_product):
        self.data_product: str                  = data_product
        self.env: str                           = Variable.get('ENV', 'dev')
        self.airflow_connection_name: str       = 'airbyte_default'
        self.airbyte_conn_id: str               = 'e6ccfe77-09d9-4829-a3fc-ce1327853337'
        self.monitoring_project: str            =  'internal-monitoring-dev-7e9d' if self.env == 'dev' else "internal-monitoring-prd-afa4"
        self.COMMON_GITLAB_PAT                  = Variable.get("COMMON_GITLAB_PAT","")
        self.GCLOUD_PROJECT = Variable.get("GCLOUD_PROJECT","")
        self.SLACK_CHANNEL_TEST = Variable.get("SLACK_CHANNEL_TEST","")
        self.SLACK_TOKEN = Variable.get("SLACK_TOKEN","")
        

        env_vars = [
            self.set_env_vars(name="STG_PROJECT_ID", key="STG_PROJECT_ID"),
            self.set_env_vars(name="LZ_PROJECT_ID", key="LZ_PROJECT_ID"),
            self.set_env_vars(name="DM_PROJECT_ID", key="DM_PROJECT_ID"),
            self.set_env_vars(name="DW_PROJECT_ID", key="DW_PROJECT_ID"),
            self.set_env_vars(name="COMMON_GITLAB_PAT", value=self.COMMON_GITLAB_PAT),
            self.set_env_vars(name="DATA_PRODUCT", value=self.data_product),
            self.set_env_vars(name="GCLOUD_PROJECT", value=self.GCLOUD_PROJECT),
            self.set_env_vars(name="SLACK_CHANNEL_TEST", value=self.SLACK_CHANNEL_TEST),
            self.set_env_vars(name="SLACK_TOKEN", value=self.SLACK_TOKEN),
        ]
        
        self.env_vars = env_vars
    
    def set_env_vars(self,name: str, value=None, key=None):
        if value:
            env_var = k8s.V1EnvVar(name=name, value=value)
            return env_var
        if key:
            env_var = k8s.V1EnvVar(name=name, value_from=k8s.V1EnvVarSource(
                config_map_key_ref=k8s.V1ConfigMapKeySelector(name="internal-data-dbt-config-map", key=key)))
            return env_var 
        
    def add_env_vars(self,additional_vars=None):
        """
        Set up environment variables for a given data product and environment,
        with the option to add additional variables.
        """

        # Extend with additional environment variables specific to the DAG
        if additional_vars:
            config_vars = copy.deepcopy(self.env_vars)
            print(additional_vars)
            config_vars.extend([
                self.set_env_vars(name=var_name, value=var_value) for var_name, var_value in additional_vars.items()
            ])

            self.env_vars  = copy.deepcopy(config_vars)
        
        return self.env_vars   
        


from airflow.providers.cncf.kubernetes.operators.pod import KubernetesPodOperator

def node_affinity(node_pool_name: object = 'dbt') -> object:
    """
    Create the affinity parameter to execute the Pod in the given node_pool_name
    Most of the executions should be run in the default node pool ('dbt').
    If the execution needs more performances, please use 'performance'.
    @param node_pool_name: the name of the node pool to execute the pod.
    """
    return {
        'nodeAffinity': {
            'requiredDuringSchedulingIgnoredDuringExecution': {
                'nodeSelectorTerms': [
                    {
                        'matchExpressions': [
                            {
                                'key': 'node_pool',
                                'operator': 'In',
                                'values': [node_pool_name]
                            }
                        ]
                    }
                ]
            }
        }
    }
def gke_bash(dag, task_id, image, arguments, env_vars, trigger_rule,affinity,service_account,cmds = ["/bin/bash", "-c"]):
    """
    Creates a pod operator with the defaults values for the Data lake.
    @param dag: The DAG object to which the task belongs.
    @param task_id: The unique identifier for the task.
    @param image: The Docker image to be used for the pod.
    @param arguments: The command to be executed inside the pod.
    @param affinity: (optional) The node affinity settings for the pod. Defaults to node_affinity().
    @return: a KubernetesPodOperator with the default values

    """
    return KubernetesPodOperator(
        image_pull_policy="Always",
        image_pull_secrets="jfrog",
        is_delete_operator_pod=True,
        startup_timeout_seconds=600,
        cmds=cmds,
        namespace='dbt',
        get_logs=True,
        trigger_rule=trigger_rule,
        service_account_name=service_account,
        env_vars=env_vars,
        affinity=affinity,
        image=image,
        task_id=task_id,
        name=task_id,
        arguments=[arguments],
        dag=dag)


#TO BE REMOVED
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))
from airflow.providers.airbyte.operators.airbyte import AirbyteTriggerSyncOperator
from airflow import DAG
from airflow.models.dag import DagContext
from astrafy_environment import AstrafyEnvironment


def airbyte_sync(task_id,env: AstrafyEnvironment,dag: DAG = None,) ->AirbyteTriggerSyncOperator:
    dag = dag or DagContext.get_current_dag()
    async_source_destination = AirbyteTriggerSyncOperator(
        task_id=task_id,
        airbyte_conn_id=env.airflow_connection_name,
        connection_id=env.airbyte_conn_id,
        asynchronous=False,
        timeout=3600,
        wait_seconds=3, dag=dag
        )
    return async_source_destination


#TO BE REMOVED
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))

def get_elementary_command(service_account, dataset,bucket,project):
    elementary_command = f"""
        mkdir -p ./profiles;
        echo \"\"\"
        elementary:
            outputs:
                default:
                    type: bigquery
                    project: {project}
                    dataset: {dataset}
                    job_execution_timeout_seconds: 300
                    job_retries: 1
                    location: EU
                    method: oauth
                    priority: interactive
                    threads: 1
                    impersonate_service_account: {service_account}
            target: default
        \"\"\" >> ./profiles/profiles.yml;
        edr send-report --gcs-bucket-name {bucket} --profiles-dir ./profiles"""
    return elementary_command


from setuptools import setup, find_packages

setup(
    name="astrafy_airflow_utils",
    version="1.0.5",
    packages=find_packages(),
    install_requires=[],
    author="Andrea Bombino",
    description="Astrafy Airflow utils packag3",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://gitlab.com/python-packages3/airflow-utils",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)


#TO BE REMOVED
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))
from airflow import DAG
from airflow.models import BaseOperator
from airflow.utils.trigger_rule import TriggerRule

from astrafy_environment import AstrafyEnvironment
from k8s_utils import node_affinity, gke_bash
  


def dbt_image(data_product:str,tag_name: str) -> str:
    """
    @return the reference to the DBT image built for the given tag_name
    @param tag_name: the name of tag
    """
    return f'europe-west1-docker.pkg.dev/prj-astrafy-artifacts/mds/dbt/{data_product}:{tag_name}'


def dbt_target_arg(environment: AstrafyEnvironment) -> str:
    return f"""--target="{environment.data_product}-{environment.env}-ci-airflow" """


def dbt_vars(other_vars: dict = None):
    """
    Builds the dbt --vars argument with the default value taken from airflow vars and environment:
    - dag_id,
    - dag_ts,
    - source_db_main
    """
    ts = "{{ ts }}"
    run_id = "{{ run_id }}"
    defaults = f""" --vars '{{ "dag_id": "{ run_id }" , "dag_ts": "{ ts }", "orchestrator": "Airflow", "job_run_id": "{run_id}" """
    if other_vars is not None:
        for key in other_vars:
            defaults = defaults + f""", "{key}": "{other_vars[key]}" """
    defaults = defaults + "}'"
    return defaults


def dbt_in_pod(dag: DAG, task_id: str, tag_version: str, data_product:str, env_vars:str,cmd: str,               
               trigger_rule: TriggerRule = TriggerRule.ONE_SUCCESS,
               affinity: object = node_affinity(),
               service_account: str = 'dbt-ksa') -> BaseOperator:
    """
    Instantiates a GKEStartPodOperator to run a dbt command. It uses all the default variables to simplify the work
    needed to create a DBT job
    @param trigger_rule: (optional) The trigger rule, default is ALL_SUCCESS
    @param affinity: the affinity to select the node pool of execution. Default to node_affinity() which use the
    'composer' node pool.
    @param dag:
    @param task_id:
    @param cmd:
    @param tag_version:
    @param service_account: if a specific service_account is needed ('dbt-ksa' by default)
    @return: the GKEStartPodOperator
    """
    return gke_bash(dag, task_id, dbt_image(data_product,tag_version), cmd, env_vars,trigger_rule, affinity,service_account)

def dbt(
        profile_arg: str,
        target_arg: str,
        dbt_command: str = "build",
        vars_arg: str = dbt_vars(),
        other_args: str = ""):
    """
    Builds the complete dbt command for seed with the upload of the dbt result files (
    manifest.json and source.json for freshness command, manifest.json and run-results.json
     for other commands) to the datalake bucket for further analysis.

    @param dbt_command: the dbt command ex: `run`, `build`, `test --select models/vc` `source freshness` defaults to `build`
    @param vars_arg: the variables of the dbt command. In the form of '--vars { "var1": "val1", "var2": "val2"}'.
                     By default, use the result of dbt_vars() which sets dag_id, dat_ts, source_db_main,
                     source_db_legacy variables. If you want more variables, you should probably call dbt_vars() with
                     some extra vars.
    @param other_args: any additional args to pass to the dbt command
    @param profile_arg: the profile arg if the default (--profiles-dir=/app/_profiles_v2) doesn't apply
    @param target_arg: the target arg if the default doesn't apply (--target="{environment.env}")
    @return: the complete command string
    """

    return f"""
    dbt {dbt_command} --profiles-dir={profile_arg} {target_arg} {vars_arg} {other_args}; ret=$?;
    
"""


def dbt_freshness_and_upload_cmd(
        profile_arg,
        target_arg: str,
        dbt_command: str = "",
        vars_arg: str = dbt_vars(),
        other_args: str = ""):
    """
    Execute the given command and upload the dbt result files (manifest.json and run-results.json)
    to the datalake bucket for further treatments.
    @param dbt_command: the arguments for the dbt run command (ex: --select models/vc )
    @param vars_arg: the variables of the dbt command. In the form of '--vars { "var1": "val1", "var2": "val2"}'.
                     By default, use the result of dbt_vars() which sets dag_id, dat_ts, source_db_main,
                     source_db_legacy variables. If you want more variables, you should probably call dbt_vars() with
                     some extra vars.
    @param other_args: any additional args to pass to the dbt command
    @param profile_arg: the profile arg if the default (--profiles-dir=/app/_profiles_v2) doesn't apply
    @param target_arg: the target arg if the default doesn't apply (--target="{environment.env}")
    @return: the complete command string
    """
    return dbt(profile_arg,target_arg,f"source freshness {dbt_command}",vars_arg,other_args)


def dbt_test_and_upload_cmd(
        profile_arg,
        target_arg: str,
        dbt_command: str = "",
        vars_arg: str = dbt_vars(),
        other_args: str = ""):
    """
    Execute the given command and upload the dbt result files (manifest.json and run-results.json)
    to the datalake bucket for further treatments.

    @deprecated use dbt_cmd() instead.
    @param dbt_command: the arguments for the dbt run command (ex: --select models/vc )
    @param vars_arg: the variables of the dbt command. In the form of '--vars { "var1": "val1", "var2": "val2"}'.
                     By default, use the result of dbt_vars() which sets dag_id, dat_ts, source_db_main,
                     source_db_legacy variables. If you want more variables, you should probably call dbt_vars() with
                     some extra vars.
    @param other_args: any additional args to pass to the dbt command
    @param profile_arg: the profile arg if the default (--profiles-dir=/app/_profiles_v2) doesn't apply
    @param target_arg: the target arg if the default doesn't apply (--target="{environment.env}")
    @return: the complete command string
    """
    return dbt(profile_arg,target_arg,f"test {dbt_command}",vars_arg,other_args)


def dbt_run_and_upload_cmd(
        profile_arg,
        target_arg: str,
        dbt_command: str = "",
        vars_arg: str = dbt_vars(),
        other_args: str = ""):
    """
    Builds the complete dbt command with the upload of the dbt result files (manifest.json and run-results.json)
    to the datalake bucket for further analysis.
    @param dbt_command: the arguments for the dbt run command (ex: --select models/vc )
    @param vars_arg: the variables of the dbt command. In the form of '--vars { "var1": "val1", "var2": "val2"}'.
                     By default, use the result of dbt_vars() which sets dag_id, dat_ts, source_db_main,
                     source_db_legacy variables. If you want more variables, you should probably call dbt_vars() with
                     some extra vars.
    @param other_args: any additional args to pass to the dbt command
    @param profile_arg: the profile arg if the default (--profiles-dir=/app/_profiles_v2) doesn't apply
    @param target_arg: the target arg if the default doesn't apply (--target="{environment.env}")
    @return: the complete command string
    """
    return dbt(profile_arg,target_arg,f"run {dbt_command}",vars_arg,other_args)


def dbt_seed_and_upload_cmd(
        profile_arg,
        target_arg: str,
        dbt_command: str = "",
        vars_arg: str = dbt_vars(),
        other_args: str = ""):
    """
    Builds the complete dbt command for seed with the upload of the dbt result files (manifest.json and run-results.json)
    to the datalake bucket for further analysis.
    @param dbt_command: the arguments for the dbt run command (ex: --select models/vc )
    @param vars_arg: the variables of the dbt command. In the form of '--vars { "var1": "val1", "var2": "val2"}'.
                     By default, use the result of dbt_vars() which sets dag_id, dat_ts, source_db_main,
                     source_db_legacy variables. If you want more variables, you should probably call dbt_vars() with
                     some extra vars.
    @param other_args: any additional args to pass to the dbt command
    @param profile_arg: the profile arg if the default (--profiles-dir=/app/_profiles_v2) doesn't apply
    @param target_arg: the target arg if the default doesn't apply (--target="{environment.env}")
    @return: the complete command string
    """
    return dbt(profile_arg,target_arg,f"seed {dbt_command}",vars_arg,other_args)
